#include "params.h"
#include "cell.h"
#include "CDF.h"
#include <vector>
#include <iostream>
#include <fstream>
#include "perf_timer.h"
using namespace std;

static void report(cellSpace* space, int tag)
{
	char filename[100];
	sprintf(filename,"s%d",tag);
	ofstream stream(filename);
	float index = 1.0;
	for (int x = 0; x < space->getWidth(); x++)
	{
		const cell* c = dynamic_cast<const cell*>(space->getModel(x));
		if (c != NULL)
		{
			adevs_vec<3> s(c->getState());
			double R = s[0];
			double U = s[1] / R;
			double P = (s[2] - 0.5 * R * U * U ) * K_1;
			unsigned long C = c->getOutputCount();
			stream <<(index-0.5)*cell::dx<<" "<<P<<" "<<R<<" "<<U<<" "<<C<<endl;
			index += 1.0;
		}
	}
	stream.close();
}

class sod_sim: public devssim
{
	public:
		sod_sim(cellSpace* space, double sample_interval):
		devssim(space),
		space(space),
		interval(sample_interval),	
		last_sample(0.0),
		index(0)
		{
			report(space,0);
		}
		bool halt()
		{
			if (timeLast() >= last_sample+interval)
			{
				report(space,index++);
				last_sample = timeLast();
			}
			return devssim::halt();
		}
	private:
		cellSpace* space;
		double interval, last_sample;
		int index;
};


static vector<adevs_vec<3> > data;
static double tend;

void Initialize( void )
{
	int LEN;
	char str[80];
	/* loading data from file "classic/godunov.ini" */
	ifstream pF ("classic/godunov.ini");
	if(pF == 0) 
	{
		cerr << "Cannot open file godunov.ini" << endl;
		exit( -1 );
   }

   int i = 0;
	double deltaT;

   do
	{
		pF.getline(str,79);
		if (!pF.eof()) 
		{
			switch(i) 
			{
				case 0: LEN    = atoi( str ); break;
				case 1: cell::dx = atof( str ); break;
				case 2: deltaT = atof( str ); break;
				case 3: tend = atof( str ); break;
				case 4: cell::D[0] = atof( str ); break;
				case 5: cell::D[1] = atof( str ); break;
				case 6: cell::D[2] = atof( str ); break;
				default: break;
			} /* end switch() */
		i++;
		} /* end if() */
		else 
		{
			cerr << "Error while reading file godunov.ini" << endl;
			exit(1);
		}
	} 
	while( i < 7 );

	pF.close();

	int LENN = LEN+1;
	adevs_vec<3> u;

	/* initial conditions */
	double U, R, P;
	for( i = 1, U = 0., R = 2., P = 200000.; i <= LEN/2; i++ ) 
	{
		u[0] = R;
		u[1] = R * U;
		u[2] = P / K_1 + 0.5 * R * U * U;
		data.push_back(u);
	} /* end for() */
   for( U = 0., R = 1., P = 100000.; i < LENN; i++ ) 
	{
		u[0] = R;
		u[1] = R * U;
		u[2] = P / K_1 + 0.5 * R * U * U;
		data.push_back(u);
	}

	/* left boundary conditions */
	u[0] = data[0].get(0);
	u[1] = -1.0 * data[0].get(1);
	u[2] = data[0].get(2);
	data.insert(data.begin(),u);
	/* right boundary conditions */
	u[0] = data[data.size()-1].get(0);
	u[1] = -1.0 * data[data.size()-1].get(1);
	u[2] = data[data.size()-1].get(2);
	data.push_back(u);
	
} /* end Initialize( ) */

static cellSpace* make_model()
{
	vector<devs*> models;
	unsigned index = 0;
	vector<adevs_vec<3> >::const_iterator i = data.begin();
	vector<adevs_vec<3> >::const_iterator k = data.begin(); k++;
	// left boundary and interior and right boundary
	for (; index < data.size()-1; index++,i++,k++)
	{
		models.push_back(new CDF(*i,*k,index==0,index==(data.size()-2)));
		if (index < data.size()-2)
		{
			models.push_back(new cell(*k));
		}
	}
	cellSpace* cells = new cellSpace(cellSpace::SIX_POINT,models.size());
	for (index = 0; index < models.size(); index++)
	{
		cells->add(models[index],index);
	}
	return cells;
}

int main(int argc, char** argv) 
{
	try 
	{
		Initialize();
		cellSpace* space = make_model();
//		sod_sim sim(space,0.1);
		devssim sim(space);
		tic();
		sim.run(tend);
		double e = tocs();
		report(space,0);
		cout << e << " " << "secs." << endl;
		cout << cell::min_ta << " was min_ta" << endl;
		delete space;
	}
	catch(adevs_exception& err) 
	{
		cerr << err.getErrorMsg() << endl;
	}
	return 0;
}


