/***************
adevs, a library supporting discrete event simulation
Copyright (C) 1999, 2000, 2001, 2002 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __adevs_pqueue_h_
#define __adevs_pqueue_h_
#include "adevs_func.h"
#include "adevs_hashtable.h"
#include "adevs_bag.h"
#define PQUEUE_DEFAULT_GROW 100
#define PQUEUE_DEFAULT_CAP 100

/**
A priority queue, implemented as a binary heap.  Priority comparison use the
<, >, >=, <=, and == operators.  Elements are sorted with the smallest at the top.
Priority types and elements types should support the assignment operator.
A hashing function must also be provided for the element types (default is to
use the id() function).
*/
template <class priority_t,class element_t,class hash_func=id<element_t> > 
class adevs_pqueue
{
	public:		
		/// Creates a heap with initial capacity and incremental growth value.
		adevs_pqueue():
		capacity(PQUEUE_DEFAULT_CAP),grow(PQUEUE_DEFAULT_GROW)
		{
			heap = new heap_element_t[capacity];
		}
		adevs_pqueue(int capacity, int grow = PQUEUE_DEFAULT_GROW):
		capacity(capacity),grow(grow)
		{
			heap = new heap_element_t[capacity];
		}
		adevs_pqueue(const adevs_pqueue<priority_t,element_t,hash_func>& src):
		members(src.members),capacity(src.capacity),grow(src.grow)
		{
			heap = new heap_element_t[capacity];
			for (int i = 0; i < capacity; i++)
				heap[i] = src.heap[i];
		}
		const adevs_pqueue<priority_t,element_t,hash_func>&
		operator=(const adevs_pqueue<priority_t,element_t,hash_func>& src)
		{
			members = src.members;
			capacity = src.capacity;
			grow = src.grow;
			delete [] heap;
			heap = new heap_element_t[capacity];
			for (int i = 0; i < capacity; i++)
				heap[i] = src.heap[i];
			return *this;
		}
		/**
		Inserts a new item into the queue.  If the item is already in the
		queue, the item is reprioritized.
		*/
		void insert(const element_t& e, const priority_t& priority)
		{
			if (!members.contains(e))
			{
				if (members.getSize()+1 == capacity)
					enlarge();
				int i = percolate_up(members.getSize()+1,priority);
				heap[i].priority = priority;
				heap[i].item = e;
				members.insert(heap[i].item,i);
			}
			else
			{
				int to_move = members[e];
				if (heap[to_move].priority > priority)
					to_move = percolate_up(to_move,priority);
				else if (heap[to_move].priority < priority)
					to_move = percolate_down(to_move,priority);
				else
					return;
				heap[to_move].priority = priority;
				heap[to_move].item = e;
				members[e] = to_move;
			}
		}
		/// Removes an arbitrary element from the queue
		void remove(const element_t& e)
		{
			int to_move = members[e];
			priority_t min_priority(minPriority());
			to_move = percolate_up(to_move,min_priority);
			heap[to_move].priority = min_priority;
			heap[to_move].item = e;
			removeMinimum();
		}
		/// Get the element at the front of the queue
		const element_t& getMinimum()
		{
			return heap[1].item;
		}
		/// Removes the element at the front of the queue.
		void removeMinimum()
		{
			heap_element_t last(heap[members.getSize()]);
			members.remove(heap[1].item);
			if (members.getSize() > 0)
			{
				int i = percolate_down(1,last.priority);
				heap[i] = last;
				members[last.item] = i;
			}
		}
		/// Returns true if the item has been enqueued, false otherwise.
		bool isEnqueued(const element_t& e)
		{
			return members.contains(e);
		}
		/**
		Returns the minimum priority.  
		*/
		const priority_t& minPriority()
		{
			return heap[1].priority;
		}
		/// Returns the number of items in the queue.
		int getSize() const 
		{ 
			return members.getSize(); 
		}
		/// Returns true if the queue is empty, false otherwise
		bool isEmpty() const
		{
			return members.getSize() == 0;
		}
		/// Removes all items in the queue
		void clear()
		{
			members.clear();
		}
		/// Destructor
		~adevs_pqueue()
		{
			delete [] heap;
		}
	private:
		// Defn of an element in the heap
		template <class _element_t,class _priority_t> class heap_element 
		{
			public:
				_element_t item;
				_priority_t priority;
				heap_element(){}
				heap_element(const _element_t& item,const _priority_t& priority):
				item(item),priority(priority){}
				heap_element(const heap_element<_element_t,_priority_t>& src):
				item(src.item),priority(src.priority){}
				const heap_element<_element_t,_priority_t>&
				operator=(const heap_element<_element_t,_priority_t>& src)
				{
					item = src.item;
					priority = src.priority;
					return *this;
				}
				~heap_element(){}
		};
		typedef heap_element<element_t,priority_t> heap_element_t;
		adevs_hashtable<element_t,int,hash_func> members;
		heap_element_t* heap;
		int capacity;
		int grow;
		
		void enlarge()
		{
			heap_element_t* rheap = new heap_element_t[capacity+grow];
			for (int i = 0; i < capacity; i++)
				rheap[i] = heap[i];
			capacity += grow;
			delete [] heap;
			heap = rheap;
		}

		int percolate_down(int index, const priority_t& priority)
		{
			int child;
			int i = index;
			int size = members.getSize();
			for (; i*2 <= size; i = child) 
			{
				child = i*2;
				if (child != size && heap[child+1].priority < heap[child].priority) 
					child++;
				if (priority > heap[child].priority) 
				{
					heap[i] = heap[child];
					members[heap[i].item] = i;
				}
				else break;
			}
			return i;
		}

		int percolate_up(int index, const priority_t& priority)
		{
			int i = index;
			while (i/2 != 0 && heap[i/2].priority >= priority) 
			{
				heap[i] = heap[i/2];
				members[heap[i].item] = i;
				i = i/2;
			}
			return i;
		}
};

#undef PQUEUE_DEFAULT_GROW 
#undef PQUEUE_DEFAULT_CAP 
#endif
