#ifndef _adevs_threads_h_
#define _adevs_threads_h_
#include "adevs_bag.h"
#include <pthread.h>
#include <semaphore.h>
/**
A condition variable controller.
*/

class cond_var
{
	public:
		cond_var()
		{
			pthread_mutex_init(&mtx,NULL);
			pthread_cond_init(&cond,NULL);
		}
		void lock()
		{
			pthread_mutex_lock(&mtx);
		}
		void unlock()
		{
			pthread_mutex_unlock(&mtx);
		}
		void wait()
		{
			pthread_cond_wait(&cond,&mtx);
		}
		void signal()
		{
			pthread_cond_signal(&cond);
		}
		~cond_var()
		{
			pthread_mutex_destroy(&mtx);
			pthread_cond_destroy(&cond);
		}
	private:
		pthread_mutex_t mtx;
		pthread_cond_t cond;

		cond_var(const cond_var& src){}	
		void operator=(const cond_var& src){}	
};

/**
Semaphore for keeping track of jobs in the job queue.
*/
class sem
{
	public:
		sem(int count = 0)
		{
			sem_init(&s,0,count);
		}
		void up()
		{
			sem_post(&s);
		}
		void down()
		{
			sem_wait(&s);
		}
		~sem()
		{
			sem_destroy(&s);
		}
	private:
		sem_t s;
	
		sem(const sem& src){}
		void operator=(const sem& src){}
};

/**
Wrapper around an OS specific thread.
*/
class thread
{
	public:
		/// Create the thread and start executing the given function.
		thread(void*(*func)(void*),void* arg)
		{
			pthread_create(&pthrd,NULL,func,arg);
			pthread_detach(pthrd);
		}
		/// Free resources associated with the thread
		~thread()
		{
		}
	private:
		pthread_t pthrd;
};

/**
A proxy for accessing a global kernel-thread pool.  
Threads from the global pool can be used to executed
application (adevs_thread) objects.
*/
class adevs_threadpool
{
	public:
		/// Create a thread pool proxy
		adevs_threadpool();
		/// Assign a job to a kernel-thread
		bool fork(void (*func)(void*,void*), void* arg1 = NULL, void* arg2 = NULL);
		/**
		Wait for all jobs submitted by this proxy to complete.
		*/
		void join_all();
		/// Execute join_all() then delete the proxy.
		~adevs_threadpool();
		/// Set an explicit upper bound on the number of threads that can be created
		static void set_thread_limit(int limit);
	private:

		/// Description of a job to be run by the thread pool
		struct job_t
		{
			/// Count of running jobs
			sem* outstanding_sem;
			/// Function to be executed and its arguments
			void (*func)(void*,void*);
			void* arg1;
			void* arg2;
		};
		// List of jobs to be completed by the thread pool
		static adevs_bag<adevs_threadpool::job_t> jobs;
		// Condition variable controlling access to the job list
		static cond_var job_cvar;
		// Flag used to terminate threads in the pool
		static bool run_threads;
		// Number of threads not assigned to jobs
		static int avail_threads;
		// Total number of threads in the pool
		static int total_threads;
		// Limit on the size of the thread pool
		static int thread_limit;
		// Semaphore for counting outstanding jobs
		sem outstanding_sem;
		int outstanding;

		void submit_job(job_t& job);
		static job_t get_job();

		friend void* run_pthread(void* arg);
		adevs_threadpool(const adevs_threadpool& src);
		void operator=(const adevs_threadpool& src);
};

#endif
