/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2003 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include "prefab_models.h"
using namespace std;

const int cellSpace::SIX_POINT = 0;
const int cellSpace::TWENTY_SIX_POINT = 1;

const port_t cell_interface::in[3][3][3] = 
{
	{ { 0 , 1, 2 }, { 3, 4, 5 }, { 6, 7, 8 } },
	{ { 9, 10, 11 }, { 12, 13, 14 }, { 15, 16, 17 } },
	{ { 18, 19, 20 }, { 21, 22, 23 }, { 24, 25, 26 } }
};
const port_t cell_interface::out = 27;

cellSpace::net_exec::net_exec(
devs* parent, int width, int height, int depth, int type):
netExec(parent),
w(width),
h(height),
d(depth),
type(type)
{
	if (w == 0 || h == 0 || d == 0)
	{
		space = NULL;
		return;
	}
	space = new devs***[w];
	for (int x = 0; x < w; x++)
	{
		space[x] = new devs**[h];
		for (int y = 0; y < h; y++)
		{
			space[x][y] = new devs*[d];
			for (int z = 0; z < d; z++)
			{
				space[x][y][z] = NULL;
			}
		}
	}
}

cellSpace::net_exec::net_exec(const net_exec& src, devs* parent):
netExec(src,parent),
w(src.w),
h(src.h),
d(src.d),
type(src.type)
{
	if (src.space == NULL)
	{
		space = NULL;
		return;
	}
	space = new devs***[w];
	for (int x = 0; x < w; x++)
	{
		space[x] = new devs**[h];
		for (int y = 0; y < h; y++)
		{
			space[x][y] = new devs*[d];
			for (int z = 0; z < d; z++)
			{
				if (src.space[x][y][z] == NULL)
				{
					space[x][y][z] = NULL;
				}
				else
				{
					space[x][y][z] = dynamic_cast<devs*>(src.space[x][y][z]->clone());
					coord_t coord(x,y,z);
					coord_table[space[x][y][z]] = coord;
				}
			}
		}
	}
}	

void cellSpace::net_exec::init()
{
	passivate();
}

void cellSpace::net_exec::getComponents(adevs_set<devs*>& c)
{
	for (int x = 0; x < w; x++)
	{
		for (int y = 0; y < h; y++)
		{
			for (int z = 0; z < d; z++)
			{
				if (space[x][y][z] != NULL)
				{
					c.add(space[x][y][z]);
				}
			}
		}
	}
}

void cellSpace::net_exec::addModel(devs* model, int x, int y, int z)
{
	coord_t coord(x,y,z);
	coord_table[model] = coord;
	space[x][y][z] = model;
}

const devs* cellSpace::net_exec::getModel(int x, int y, int z) const
{
	return space[x][y][z];
}

devs* cellSpace::net_exec::getModel(port_t port)
{
	devs* model = NULL;
	memcpy((char*)&model,(const char*)&port,sizeof(port_t));
	return model;
}

void cellSpace::net_exec::route(
const PortValue& pv, devs* model, adevs_bag<EventReceiver>& r)
{
	internal_coupling(pv,model,r);
	if (getParent() != NULL)
	{
		external_input_coupling(pv,model,r);
		external_output_coupling(pv,model,r);
	}
}

void cellSpace::net_exec::external_input_coupling(
const PortValue& pv, devs* model, adevs_bag<EventReceiver>& r)
{
	if (model == getParent())
	{
		EventReceiver e(getModel(pv.port),cell_interface::in[1][1][1]);
		r.add(e);
	}
}

void cellSpace::net_exec::external_output_coupling(
const PortValue& pv, devs* model, adevs_bag<EventReceiver>& r)
{
	if (model != getParent())
	{
		EventReceiver e(getParent(),getPort(model));
		r.add(e);
	}
}

void cellSpace::net_exec::internal_coupling(
const PortValue& pv, devs* model, adevs_bag<EventReceiver>& r)
{
	coord_t pos(coord_table[model]);
	EventReceiver e;
	if (type == SIX_POINT)
	{
		/// Send event left
		if (pos.x > 0)
		{
			e.model = space[pos.x-1][pos.y][pos.z];
			e.port = cell_interface::in[2][1][1];
			if (e.model != NULL) r.add(e);
		}
		/// Send event right
		if (pos.x < w-1)
		{
			e.model = space[pos.x+1][pos.y][pos.z];
			e.port = cell_interface::in[0][1][1];
			if (e.model != NULL) r.add(e);
		}
		/// Send event back
		if (pos.y > 0)
		{
			e.model = space[pos.x][pos.y-1][pos.z];
			e.port = cell_interface::in[1][2][1];
			if (e.model != NULL) r.add(e);
		}
		/// Send event forward
		if (pos.y < h-1)
		{
			e.model = space[pos.x][pos.y+1][pos.z];
			e.port = cell_interface::in[1][0][1];
			if (e.model != NULL) r.add(e);
		}
		/// Send event down
		if (pos.z > 0)
		{
			e.model = space[pos.x][pos.y][pos.z-1];
			e.port = cell_interface::in[1][1][2];
			if (e.model != NULL) r.add(e);
		}
		/// Send event up
		if (pos.z < d-1)
		{
			e.model = space[pos.x][pos.y][pos.z+1];
			e.port = cell_interface::in[1][1][0];
			if (e.model != NULL) r.add(e);
		}
	}
	else if (type == TWENTY_SIX_POINT)
	{
		for (int x = -1; x < 2; x++)
		{
			for (int y = -1; y < 2; y++)
			{
				for (int z = -1; z < 2; z++)
				{
					/// Get coords of neighboring cell
					int xx = pos.x + x;
					int yy = pos.y + y;
					int zz = pos.z + z;
					/// Check that coord is valid
					if (xx >= 0 && xx < w && yy >= 0 && yy < h && zz >= 0 && zz < d)
					{
						e.model = space[xx][yy][zz];
						if (e.model != model && e.model != NULL)
						{
							e.port = cell_interface::in[1-x][1-y][1-z];
							r.add(e);
						}
					}
				}
			}
		}
	}	
}

cellSpace::coord_t cellSpace::net_exec::getCoordinates(port_t port) const
{
	devs* pntr = NULL;
	memcpy((char*)&pntr,(const char*)&port,sizeof(port_t));
	return ((*(coord_table.find(pntr))).second);
}

port_t cellSpace::net_exec::getPort(int x, int y, int z) const
{
	port_t port;
	memcpy((char*)&port,(const char*)&(space[x][y][z]),sizeof(port_t));
	return port;
}

port_t cellSpace::net_exec::getPort(const devs* model) const
{
	port_t port;
	memcpy((char*)&port,(const char*)&model,sizeof(port_t));
	return port;
}

int cellSpace::net_exec::getDistance(devs* model)
{
	coord_t p((*(coord_table.find(model))).second);
	int result = p.x;
	if (h > w && h > d) result = p.y;
	else if (d > w && d > h) result = p.z;
	return result;
}

cellSpace::net_exec::~net_exec()
{
	for (int x = 0; x < w; x++)
	{
		for (int y = 0; y < h; y++)
		{
			for (int z = 0; z < d; z++)
			{
				if (space[x][y][z] != NULL)
				{
					delete space[x][y][z];
				}
			}
			delete [] space[x][y];
		}
		delete [] space[x];
	}
	delete [] space;
}
